﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using TittyMagic.Handlers.Configs;
using TittyMagic.Models;
using UnityEngine;
using static TittyMagic.Handlers.HandlerUtils;
using static TittyMagic.ParamName;
using static TittyMagic.Script;
using static TittyMagic.Side;

namespace TittyMagic.Handlers
{
    public static class MainPhysicsHandler
    {
        public static Rigidbody chestRb { get; set; }
        public static AdjustJoints breastControl { get; set; }

        private static Dictionary<string, ConfigurableJoint> _joints;
        private static Dictionary<string, Rigidbody> _pectoralRbs;
        private static Dictionary<string, DAZBone> _dazBones;

        private static List<string> _breastControlFloatParamNames;

        // AdjustJoints.joint1.slerpDrive.maximumForce value logged on plugin Init
        private const float DEFAULT_SLERP_MAX_FORCE = 500;

        public static Dictionary<string, PhysicsParameterGroup> parameterGroups { get; private set; }

        public static float realMassAmount { get; private set; }
        public static float massAmount { get; private set; }

        // hack. 1.5f because 3f is the max mass and massValue is actual mass / 2
        public static float InvertMass(float x) => 1.5f - x;
        public static float NormalizeMass(float x) => Mathf.InverseLerp(0, 1.45f, x);
        public static float normalizedMass => NormalizeMass(massAmount);
        public static float normalizedInvertedMass => NormalizeMass(InvertMass(massAmount));
        public static float normalizedInvertedRealMass => NormalizeMass(InvertMass(realMassAmount));

        public static MassParameterGroup massParameterGroup { get; private set; }

        private static bool _initialized;

        public static void Init()
        {
            _joints = new Dictionary<string, ConfigurableJoint>
            {
                { LEFT, breastControl.joint2 },
                { RIGHT, breastControl.joint1 },
            };
            _pectoralRbs = new Dictionary<string, Rigidbody>
            {
                { LEFT, _joints[LEFT].GetComponent<Rigidbody>() },
                { RIGHT, _joints[RIGHT].GetComponent<Rigidbody>() },
            };
            _dazBones = new Dictionary<string, DAZBone>
            {
                { LEFT, _joints[LEFT].GetComponent<DAZBone>() },
                { RIGHT, _joints[RIGHT].GetComponent<DAZBone>() },
            };
            _rotationZZeroCount = new Dictionary<string, int>
            {
                { LEFT, 0 },
                { RIGHT, 0 },
            };

            _breastControlFloatParamNames = new List<string>
            {
                MASS,
                CENTER_OF_GRAVITY_PERCENT,
                SPRING,
                DAMPER,
                POSITION_SPRING_Z,
                POSITION_DAMPER_Z,
                TARGET_ROTATION_X,
                TARGET_ROTATION_Y,
                TARGET_ROTATION_Z,
            };

            _initialized = true;
        }

        public static void UpdateMassValueAndAmounts()
        {
            float leftVolume = CalculateVolume(VertexIndexGroup.leftBreast);
            float rightVolume = CalculateVolume(VertexIndexGroup.rightBreast);
            float volume = (leftVolume + rightVolume) / 2;
            massParameterGroup.UpdateValue(volume / 1000);

            /* Division by 2 is a hacky way to make the value compatible with legacy configurations for morphs and physics settings */
            realMassAmount = massParameterGroup.left.baseValueJsf.val / 2;
            massAmount = massParameterGroup.left.valueJsf.val / 2;
        }

        private static float CalculateVolume(int[] vertexIndices)
        {
            var positions = new Vector3[vertexIndices.Length];
            for(int i = 0; i < vertexIndices.Length; i++)
            {
                positions[i] = Calc.RelativePosition(chestRb, skin.rawSkinnedVerts[vertexIndices[i]]);
            }

            var bounds = new Bounds();

            /* Calculate bounds size */
            {
                var min = Vector3.one * float.MaxValue;
                var max = Vector3.one * float.MinValue;
                foreach(var vertex in positions)
                {
                    min = Vector3.Min(min, vertex);
                    max = Vector3.Max(max, vertex);
                }

                bounds.min = min;
                bounds.max = max;
            }

            /* Calculate volume */
            {
                float toCm3 = Mathf.Pow(10, 6);
                float scale = tittyMagic.scaleJsf.val;

                /* This somewhat accurately scales breast volume to the apparent breast size when atom scale is adjusted. */
                float atomScaleAdjustment = 1 - Mathf.Abs(Mathf.Log10(Mathf.Pow(scale, 3)));
                float atomScaleFactor = scale >= 1
                    ? scale * atomScaleAdjustment
                    : scale / atomScaleAdjustment;

                float z = bounds.size.z * atomScaleFactor;
                float volume = toCm3 * (4 * Mathf.PI * bounds.size.x / 2 * bounds.size.y / 2 * z / 2) / 3;

                /* Times 0.75f compensates for change in estimated volume compared to pre v3.2 bounds calculation */
                return volume * 0.75f;
            }
        }

        public static void LoadSettings()
        {
            SetupPhysicsParameterGroups();

            var texts = CreateInfoTexts();
            massParameterGroup.infoText = texts[MASS];
            foreach(var param in parameterGroups)
            {
                param.Value.infoText = texts[param.Key];
            }
        }

        #region *** Parameter setup ***

        private static MassParameter NewMassParameter(string side)
        {
            string jsfName = JsfName(MASS, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0.100f, 0.100f, 3.000f);
            return new MassParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                valueFormat = "F3",
                sync = value => SyncMass(_pectoralRbs[side], value),
            };
        }

        private static PhysicsParameter NewCenterOfGravityParameter(string side)
        {
            string jsfName = JsfName(CENTER_OF_GRAVITY_PERCENT, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0.00f, 0.00f, 1.00f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 0.60f,
                    massCurve = x => 0.25f * x,
                    softnessCurve = x => 0.08f * x,
                },
                valueFormat = "F2",
                sync = value => SyncCenterOfGravity(_pectoralRbs[side], value),
            };
        }

        private static PhysicsParameter NewSpringParameter(string side)
        {
            string jsfName = JsfName(SPRING, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 10f, 10f, 100f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 57f,
                    massCurve = x => 0.10f * x,
                    // https://www.desmos.com/calculator/nxyosar9o6
                    softnessCurve = x => -0.50f * Curves.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                quicknessOffsetConfig = new StaticPhysicsConfig
                {
                    baseValue = 24f,
                },
                slownessOffsetConfig = new StaticPhysicsConfig
                {
                    baseValue = -12f,
                },
                valueFormat = "F0",
                sync = value => SyncJointSpring(_joints[side], _pectoralRbs[side], value),
            };
        }

        private static PhysicsParameter NewDamperParameter(string side)
        {
            string jsfName = JsfName(DAMPER, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0.00f, 0.00f, 10.00f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 1.28f,
                    // https://www.desmos.com/calculator/y3akvzgr1s
                    massCurve = x => 1.00f * Curves.InverseSmoothStep(2 / 3f * x, 1.00f, 0.30f, 0.60f),
                    // https://www.desmos.com/calculator/nxyosar9o6
                    softnessCurve = x => -0.60f * Curves.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                altConfig = new StaticPhysicsConfig
                {
                    baseValue = 1.10f,
                    // https://www.desmos.com/calculator/y3akvzgr1s
                    massCurve = x => 1.35f * Curves.InverseSmoothStep(2 / 3f * x, 1.00f, 0.30f, 0.60f),
                    // https://www.desmos.com/calculator/nxyosar9o6
                    softnessCurve = x => -0.80f * Curves.InverseSmoothStep(x, 1.00f, 0.24f, 0.61f),
                },
                valueFormat = "F2",
                sync = value => SyncJointDamper(_joints[side], _pectoralRbs[side], value),
            };
        }

        private static PhysicsParameter NewPositionSpringZParameter(string side)
        {
            string jsfName = JsfName(POSITION_SPRING_Z, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0f, 0f, 1000f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 720f,
                    massCurve = x => -0.14f * InvertMass(x),
                    softnessCurve = x => -0.45f * x,
                },
                valueFormat = "F0",
                sync = value => SyncJointPositionZDriveSpring(_joints[side], _pectoralRbs[side], value),
            };
        }

        private static PhysicsParameter NewPositionDamperZParameter(string side)
        {
            string jsfName = JsfName(POSITION_DAMPER_Z, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0f, 0f, 100f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 11f,
                },
                valueFormat = "F0",
                sync = value => SyncJointPositionZDriveDamper(_joints[side], _pectoralRbs[side], value),
            };
        }

        private static PhysicsParameter NewTargetRotationYParameter(string side)
        {
            string jsfName = JsfName(TARGET_ROTATION_Y, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0.00f, -20.00f, 20.00f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 0.00f,
                },
                valueFormat = "F2",
            };
        }

        private static PhysicsParameter NewTargetRotationXParameter(string side)
        {
            string jsfName = JsfName(TARGET_ROTATION_X, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0.00f, -20.00f, 20.00f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, side == LEFT),
                config = new StaticPhysicsConfig
                {
                    baseValue = 0.00f,
                },
                valueFormat = "F2",
            };
        }

        private static PhysicsParameter NewTargetRotationZParameter(string side)
        {
            string jsfName = JsfName(TARGET_ROTATION_Z, side);
            var valueJsf = new JSONStorableFloat($"{jsfName}Value", 0.00f, -30.00f, 30.00f);
            return new PhysicsParameter
            {
                valueJsf = valueJsf,
                baseValueJsf = NewBaseValueJsf(jsfName, valueJsf),
                offsetJsf = NewOffsetJsf(jsfName, valueJsf, false),
                config = new StaticPhysicsConfig
                {
                    baseValue = 0.00f,
                },
                valueFormat = "F2",
            };
        }

        private static void SetupPhysicsParameterGroups()
        {
            massParameterGroup = new MassParameterGroup
            {
                displayName = "Breast Mass",
                left = NewMassParameter(LEFT),
                right = NewMassParameter(RIGHT),
                requiresCalibration = true,
            };

            var centerOfGravityPercent = new PhysicsParameterGroup
            {
                displayName = "Center Of Gravity",
                left = NewCenterOfGravityParameter(LEFT),
                right = NewCenterOfGravityParameter(RIGHT),
                requiresCalibration = true,
            };

            var spring = new PhysicsParameterGroup
            {
                displayName = "Spring",
                left = NewSpringParameter(LEFT),
                right = NewSpringParameter(RIGHT),
                requiresCalibration = true,
            };

            var damper = new PhysicsParameterGroup
            {
                displayName = "Damper",
                left = NewDamperParameter(LEFT),
                right = NewDamperParameter(RIGHT),
                dependsOnPhysicsRate = true,
            };

            var positionSpringZ = new PhysicsParameterGroup
            {
                displayName = "In/Out Spring",
                left = NewPositionSpringZParameter(LEFT),
                right = NewPositionSpringZParameter(RIGHT),
                requiresCalibration = true,
            };

            var positionDamperZ = new PhysicsParameterGroup
            {
                displayName = "In/Out Damper",
                left = NewPositionDamperZParameter(LEFT),
                right = NewPositionDamperZParameter(RIGHT),
                dependsOnPhysicsRate = true,
            };

            var leftXParam = NewTargetRotationXParameter(LEFT);
            var leftYParam = NewTargetRotationYParameter(LEFT);
            var leftZParam = NewTargetRotationZParameter(LEFT);
            leftXParam.sync = value => SyncTargetRotation(LEFT, value, leftYParam.valueJsf.val, leftZParam.valueJsf.val);
            //y and z have no sync callback, handled by x sync as a side effect

            var rightXParam = NewTargetRotationXParameter(RIGHT);
            var rightYParam = NewTargetRotationYParameter(RIGHT);
            var rightZParam = NewTargetRotationZParameter(RIGHT);
            rightXParam.sync = value => SyncTargetRotation(RIGHT, value, rightYParam.valueJsf.val, rightZParam.valueJsf.val);
            //y and z have no sync callback, handled by x sync as a side effect

            var targetRotationX = new PhysicsParameterGroup
            {
                displayName = "Up/Down Angle Target",
                left = leftXParam,
                right = rightXParam,
                requiresCalibration = true,
            };

            var targetRotationY = new PhysicsParameterGroup
            {
                displayName = "Right/Left Angle Target",
                left = leftYParam,
                right = rightYParam,
                requiresCalibration = true,
                rightInverted = true,
            };

            var targetRotationZ = new PhysicsParameterGroup
            {
                displayName = "Twist Angle Target",
                left = leftZParam,
                right = rightZParam,
                requiresCalibration = true,
                rightInverted = true, // correct but has no effect because offset not adjusted directly
                allowOffsetOnlyLeftBreast = false,
            };

            massParameterGroup.SetOffsetCallbackFunctions();
            centerOfGravityPercent.SetOffsetCallbackFunctions();
            spring.SetOffsetCallbackFunctions();
            damper.SetOffsetCallbackFunctions();
            positionSpringZ.SetOffsetCallbackFunctions();
            positionDamperZ.SetOffsetCallbackFunctions();
            targetRotationY.SetOffsetCallbackFunctions();
            targetRotationX.SetOffsetCallbackFunctions();
            targetRotationZ.SetOffsetCallbackFunctions();

            parameterGroups = new Dictionary<string, PhysicsParameterGroup>
            {
                { CENTER_OF_GRAVITY_PERCENT, centerOfGravityPercent },
                { SPRING, spring },
                { DAMPER, damper },
                { POSITION_SPRING_Z, positionSpringZ },
                { POSITION_DAMPER_Z, positionDamperZ },
                { TARGET_ROTATION_X, targetRotationX },
                { TARGET_ROTATION_Y, targetRotationY },
                { TARGET_ROTATION_Z, targetRotationZ },
            };
        }

        #endregion *** Parameter setup ***

        #region *** Sync functions ***

        // Reimplements AdjustJoints.cs method SyncMass
        private static void SyncMass(Rigidbody rb, float value)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            if(Math.Abs(rb.mass - value) <= 0.001f)
            {
                return;
            }

            rb.mass = value;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncCenterOfGravity
        private static void SyncCenterOfGravity(Rigidbody rb, float value)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            var newCenterOfMass = Vector3.Lerp(breastControl.lowCenterOfGravity, breastControl.highCenterOfGravity, value);
            if(Calc.VectorIsEqualWithin(1 / 100f, rb.centerOfMass, newCenterOfMass))
            {
                return;
            }

            rb.centerOfMass = newCenterOfMass;
            rb.WakeUp();
        }

        private static void SyncJointSpring(ConfigurableJoint joint, Rigidbody rb, float spring)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, breastControl.scale - 1f);

            float scaledSpring = spring * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionSpring - scaledSpring) <= 1)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionSpring = scaledSpring;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionSpring = scaledSpring;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionSpring = scaledSpring;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.spring = scaledSpring * breastControl.limitSpringMultiplier;
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.spring = scaledSpring * breastControl.limitSpringMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        private static void SyncJointDamper(ConfigurableJoint joint, Rigidbody rb, float damper)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            // see AdjustJoints.cs method ScaleChanged
            float scalePow = Mathf.Pow(1.7f, breastControl.scale - 1f);

            float scaledDamper = damper * scalePow;
            if(Mathf.Abs(joint.slerpDrive.positionDamper - scaledDamper) <= 0.01f)
            {
                return;
            }

            var slerpDrive = joint.slerpDrive;
            slerpDrive.positionDamper = scaledDamper;
            slerpDrive.maximumForce = DEFAULT_SLERP_MAX_FORCE * scalePow;
            joint.slerpDrive = slerpDrive;

            var angularXDrive = joint.angularXDrive;
            angularXDrive.positionDamper = scaledDamper;
            joint.angularXDrive = angularXDrive;

            var angularYZDrive = joint.angularYZDrive;
            angularYZDrive.positionDamper = scaledDamper;
            joint.angularYZDrive = angularYZDrive;

            var angularXLimitSpring = joint.angularXLimitSpring;
            angularXLimitSpring.damper = scaledDamper * breastControl.limitDamperMultiplier;
            joint.angularXLimitSpring = angularXLimitSpring;

            var angularYZLimitSpring = joint.angularYZLimitSpring;
            angularYZLimitSpring.damper = scaledDamper * breastControl.limitDamperMultiplier;
            joint.angularYZLimitSpring = angularYZLimitSpring;

            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        private static void SyncJointPositionZDriveSpring(ConfigurableJoint joint, Rigidbody rb, float spring)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            var zDrive = joint.zDrive;
            if(Mathf.Abs(zDrive.positionSpring - spring) <= 1)
            {
                return;
            }

            zDrive.positionSpring = spring;
            joint.zDrive = zDrive;
            rb.WakeUp();
        }

        // Reimplements AdjustJoints.cs method SyncJointPositionZDrive
        private static void SyncJointPositionZDriveDamper(ConfigurableJoint joint, Rigidbody rb, float damper)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            var zDrive = joint.zDrive;
            if(Mathf.Abs(zDrive.positionDamper - damper) <= 1)
            {
                return;
            }

            zDrive.positionDamper = damper;
            joint.zDrive = zDrive;
            rb.WakeUp();
        }

        private static Dictionary<string, int> _rotationZZeroCount;

        private static void SyncTargetRotation(string side, float targetRotationX, float targetRotationY, float targetRotationZ)
        {
            if(!tittyMagic.enabled)
            {
                return;
            }

            var rotation = side == LEFT
                ? breastControl.smoothedJoint2TargetRotation
                : breastControl.smoothedJoint1TargetRotation;
            rotation.x = -targetRotationX;
            rotation.y = -targetRotationY;

            bool noTargetRotationZ = GravityPhysicsHandler.targetRotationZJsf.val == 0;
            if(_rotationZZeroCount[side] < 3)
            {
                rotation.z = targetRotationZ;
                if(noTargetRotationZ)
                {
                    _rotationZZeroCount[side]++;
                }
            }
            else if(!noTargetRotationZ)
            {
                _rotationZZeroCount[side] = 0;
            }

            _dazBones[side].baseJointRotation = rotation;
        }

        #endregion *** Sync functions ***

        public static void UpdatePhysics()
        {
            float softness = tittyMagic.softnessAmount;
            float quickness = tittyMagic.quicknessAmount;
            parameterGroups.Values
                .ToList()
                .ForEach(paramGroup => UpdateParam(paramGroup, softness, quickness));
        }

        public static void UpdateRateDependentPhysics()
        {
            float softness = tittyMagic.softnessAmount;
            float quickness = tittyMagic.quicknessAmount;
            parameterGroups.Values
                .Where(paramGroup => paramGroup.dependsOnPhysicsRate)
                .ToList()
                .ForEach(paramGroup => UpdateParam(paramGroup, softness, quickness));
        }

        private static void UpdateParam(PhysicsParameterGroup paramGroup, float softness, float quickness)
        {
            float massValue = paramGroup.usesRealMass ? realMassAmount : massAmount;
            paramGroup.UpdateValue(massValue, softness, quickness);
        }

        public static void RestoreOriginalPhysics()
        {
            if(!_initialized)
            {
                return;
            }

            foreach(string name in _breastControlFloatParamNames)
            {
                /* Set a value that is different from the original, then restore the original
                 * in order to trigger VaM's internal sync
                 */
                var paramJsf = breastControl.GetFloatJSONParam(name);
                float original = paramJsf.val;
                paramJsf.valNoCallback = Math.Abs(paramJsf.val - paramJsf.min) > 0.01f
                    ? paramJsf.min
                    : paramJsf.max;
                paramJsf.val = original;
            }
        }

        private static Dictionary<string, string> CreateInfoTexts()
        {
            string massText;
            {
                var sb = new StringBuilder();
                sb.Append("Mass of the pectoral joint.");
                sb.Append("\n\n");
                sb.Append("Since mass represents breast size, other physics parameters are adjusted based on its value.");
                if(personIsFemale)
                {
                    sb.Append("\n\n");
                    sb.Append("Fat Collider Radius and Fat Distance Limit are adjusted using the mass estimated from");
                    sb.Append(" volume, the rest are adjusted using the actual mass value that includes the offset.");
                }

                massText = sb.ToString();
            }

            string centerOfGravityText;
            {
                var sb = new StringBuilder();
                sb.Append("Position of the pectoral joint's center of mass.");
                sb.Append("\n\n");
                sb.Append("At 0, the center of mass is inside the chest at the pectoral joint. At 1, it is at the nipple.");
                sb.Append(" Between about 0.5 and 0.8, the center of mass is within the bulk of the breast volume.");
                centerOfGravityText = sb.ToString();
            }

            string springText;
            {
                var sb = new StringBuilder();
                sb.Append("Magnitude of the spring that pushes the pectoral joint towards its angle target.");
                sb.Append("\n\n");
                sb.Append("The angle target is defined by the Up/Down, Left/Right and Twist Angle Target parameters.");
                springText = sb.ToString();
            }

            string damperText;
            {
                var sb = new StringBuilder();
                sb.Append("Magnitude of the damper that reduces oscillation around the joint angle target.");
                sb.Append("\n\n");
                sb.Append("The higher the damper, the quicker breasts will stop swinging.");
                damperText = sb.ToString();
            }

            string positionSpringZText;
            {
                var sb = new StringBuilder();
                sb.Append("Magnitude of the spring that pushes the pectoral joint towards its position target along the Z axis.");
                sb.Append("\n\n");
                sb.Append("Directional force morphing along the forward-back axis depends on In/Out Spring being suitably low");
                sb.Append(" for the given breast mass.");
                positionSpringZText = sb.ToString();
            }

            string positionDamperZText;
            {
                var sb = new StringBuilder();
                sb.Append("Magnitude of the damper that reduces oscillation around the joint position target along the Z axis.");
                positionDamperZText = sb.ToString();
            }

            string targetRotationXText;
            {
                var sb = new StringBuilder();
                sb.Append("Vertical target angle of the pectoral joint.");
                sb.Append(" Negative values pull breasts down, positive values push them up.");
                sb.Append("\n\n");
                sb.Append("The offset shifts the center around which the final angle is calculated");
                sb.Append(" based on the person's pose (see Gravity Multipliers).");
                targetRotationXText = sb.ToString();
            }

            string targetRotationYText;
            {
                var sb = new StringBuilder();
                sb.Append("Horizontal target angle of the pectoral joint.");
                sb.Append("\n\n");
                sb.Append("A negative offset pulls breasts apart, while a positive offset pushes them together.");
                targetRotationYText = sb.ToString();
            }

            string targetRotationZtext;
            {
                var sb = new StringBuilder();
                sb.Append("Forward axis target angle of the pectoral joint.");
                sb.Append("\n\n");
                sb.Append("The final value depends on the person's pose. The offset determines the max angle");
                sb.Append(" when the person is upright. The angle is inverted when upside down and zero when");
                sb.Append(" horizontal.");
                targetRotationZtext = sb.ToString();
            }

            return new Dictionary<string, string>
            {
                { MASS, massText },
                { CENTER_OF_GRAVITY_PERCENT, centerOfGravityText },
                { SPRING, springText },
                { DAMPER, damperText },
                { POSITION_SPRING_Z, positionSpringZText },
                { POSITION_DAMPER_Z, positionDamperZText },
                { TARGET_ROTATION_X, targetRotationXText },
                { TARGET_ROTATION_Y, targetRotationYText },
                { TARGET_ROTATION_Z, targetRotationZtext },
            };
        }

        public static void Destroy()
        {
            chestRb = null;
            breastControl = null;
            _joints = null;
            _pectoralRbs = null;
            _dazBones = null;
            _breastControlFloatParamNames = null;
            parameterGroups = null;
            massParameterGroup = null;
        }
    }
}
